@@ -57,6 +57,7 @@
 
 #include "src/core/lib/address_utils/sockaddr_utils.h"
 #include "src/core/lib/event_engine/default_event_engine.h"
+#include "src/core/lib/event_engine/memory_allocator_factory.h"
 #include "src/core/lib/event_engine/resolved_address_internal.h"
 #include "src/core/lib/event_engine/shim.h"
 #include "src/core/lib/event_engine/tcp_socket_utils.h"
@@ -85,24 +86,10 @@ static constexpr grpc_core::Duration kRetryAcceptWaitTime{
 using ::grpc_event_engine::experimental::EndpointConfig;
 using ::grpc_event_engine::experimental::EventEngine;
 using ::grpc_event_engine::experimental::MemoryAllocator;
+using ::grpc_event_engine::experimental::MemoryQuotaBasedMemoryAllocatorFactory;
 using ::grpc_event_engine::experimental::PosixEventEngineWithFdSupport;
 using ::grpc_event_engine::experimental::SliceBuffer;
 
-class MemoryAllocatorFactoryWrapper
-    : public grpc_event_engine::experimental::MemoryAllocatorFactory {
- public:
-  explicit MemoryAllocatorFactoryWrapper(
-      grpc_core::MemoryQuotaRefPtr memory_quota)
-      : memory_quota_(std::move(memory_quota)) {}
-
-  MemoryAllocator CreateMemoryAllocator(absl::string_view name) override {
-    return memory_quota_->CreateMemoryAllocator(name);
-  }
-
- private:
-  grpc_core::MemoryQuotaRefPtr memory_quota_;
-};
-
 static grpc_error_handle CreateEventEngineListener(
     grpc_tcp_server* s, grpc_closure* shutdown_complete,
     const EndpointConfig& config, grpc_tcp_server** server) {
@@ -196,7 +183,9 @@ static grpc_error_handle CreateEventEngineListener(
         delete s->fd_handler;
         delete s;
       },
-      config, std::make_unique<MemoryAllocatorFactoryWrapper>(s->memory_quota));
+      config,
+      std::make_unique<MemoryQuotaBasedMemoryAllocatorFactory>(
+          s->memory_quota));
   if (!listener.ok()) {
     delete s;
     *server = nullptr;
@@ -366,7 +355,7 @@ static void on_read(void* arg, grpc_error_handle err) {
       // This is not a performant code path, but if an fd limit has been
       // reached, the system is likely in an unhappy state regardless.
       if (errno == EMFILE) {
-        GRPC_LOG_EVERY_N_SEC(1, "%s",
+        GRPC_LOG_EVERY_N_SEC(1, GPR_ERROR, "%s",
                              "File descriptor limit reached. Retrying.");
         grpc_fd_notify_on_read(sp->emfd, &sp->read_closure);
         if (gpr_atm_full_xchg(&sp->retry_timer_armed, true)) return;
